#include "stereo.h"


inline uint64 getCensus(uchar* source, int width, int i, int j)
{
    // Census
    uint64 value = 0;
    uchar center = source[i*width + j];

    for (int m = -3; m <= 3; m++)
    {
        for (int n =  -4; n <= 4; n++)
        {
            value = (value << 1);
            value += (source[(i+m)*width+j+n] < center);
        }
    }

    return value;
}


uchar hamDist16(ushort x, ushort y)
{
    uchar dist = 0;
    ushort val = (x^y); // yihuo

    while(val)
    {
        ++dist;
        val &= val - 1;
    }

    return dist;
}

void Census(uchar *left, uchar *right, int height, int width, int dispnum, u16 *dsi)
{
    static bool firstRun = true;

    static uchar *popcount16LUT = new uchar[65536];
    uint64 *leftCensus   = new uint64[width];
    uint64 *rightCensus = new uint64[width];

    // system init
    if(firstRun)
    {
        // build the popcount16LUT
        for (int i = 0; i < 65536; i++) {
            popcount16LUT[i] = hamDist16(i,0);
        }

        memset(dsi,0,height*width*dispnum*sizeof(u16));
        firstRun = false;
    }

    for (int i = 3; i < height - 3; i++)
    {
        // Census
        for (int j = 4; j < width - 4; j++)
        {
            leftCensus[j]   = getCensus(left, width, i, j);
            rightCensus[j] = getCensus(right, width, i, j);
        }

        // Hamming Distance
        for (int j = 4; j < width - 4; j++)
        {
            for (int d = 0; d < dispnum; d++)
            {
                if( d <= j - 4)
                {
                    // total 44 bytes
                    dsi[(i*width+j)*dispnum + d] = popcount16LUT[ (ushort)(leftCensus[j] ^ rightCensus[j-d]) ]
                            + popcount16LUT[ (ushort)((leftCensus[j]>>16) ^ (rightCensus[j-d]>>16)) ]
                            + popcount16LUT[ (ushort)((leftCensus[j]>>32) ^ (rightCensus[j-d]>>32)) ];
                }
                else
                {
                    dsi[(i*width+j)*dispnum + d] = 0xff;
                }
            }
        }
    }

    delete [] leftCensus;
    delete [] rightCensus;
}


void TransposeDSI(int width, int height, int n_disp, u16 *dsiSrc, u16 *dsiDes)
{
    int index1, index2;
    for(int j=0;j<height;j++)
    {
        for(int i=0;i<width;i++)
        {
            index1=j*width*n_disp+i*n_disp;
            index2=i*height*n_disp+j*n_disp;
            memcpy(dsiDes+index2, dsiSrc+index1, sizeof(u16)*n_disp);
        }
    }
}
